import type { APIRequestContext } from '@playwright/test';
import FastGlob from "fast-glob";
import { createReadStream } from 'fs';
import { basename, resolve } from 'path';
import * as urls from '../utilities/urls.js';
import gql from './gql.js';
import time from "./time.js";

/**
 * Aerie API request functions.
 */
const req = {
  async createMissionModel(request: APIRequestContext, model: MissionModelInsertInput): Promise<number> {
    const data = await req.hasura(request, gql.CREATE_MISSION_MODEL, { model: model });
    const { insert_mission_model_one } = data;
    const { id: mission_model_id } = insert_mission_model_one;

    return mission_model_id;
  },

  async deleteMissionModel(request: APIRequestContext, id: number): Promise<number> {
    const data = await req.hasura(request, gql.DELETE_MISSION_MODEL, { id: id });
    const { delete_mission_model_by_pk } = data;
    const { id: deleted_mission_model_id } = delete_mission_model_by_pk;
    return deleted_mission_model_id;
  },

  async hasura<T = any>(
    request: APIRequestContext,
    query: string,
    variables: Record<string, unknown> = {},
  ): Promise<T> {
    const options = { data: { query, variables } };
    const response = await request.post(`${urls.HASURA_URL}/v1/graphql`, options);

    if (response.ok()) {
      const json = await response.json();

      if (json?.data) {
        const { data } = json;
        return data as T;
      } else if (json?.errors) {
        console.error(json.errors);
        const [{ message }] = json.errors;
        throw new Error(message);
      } else {
        throw new Error('An unexpected error ocurred');
      }
    } else {
      throw new Error(response.statusText());
    }
  },

  async healthGateway(request: APIRequestContext): Promise<boolean> {
    const response = await request.get(`${urls.GATEWAY_URL}/health`);
    return response.ok();
  },

  async healthHasura(request: APIRequestContext): Promise<boolean> {
    const response = await request.get(`${urls.HASURA_URL}/healthz`);
    return response.ok();
  },

  async healthUI(request: APIRequestContext): Promise<boolean> {
    const response = await request.get(`${urls.UI_URL}/health`);
    return response.ok();
  },

  async healthMerlin(request: APIRequestContext): Promise<boolean> {
    const response = await request.get(`${urls.MERLIN_URL}/health`);
    return response.ok();
  },

  async healthCommanding(request: APIRequestContext): Promise<boolean> {
    const response = await request.get(`${urls.COMMANDING_URL}/health`);
    return response.ok();
  },

  async healthScheduler(request: APIRequestContext): Promise<boolean> {
    const response = await request.get(`${urls.SCHEDULER_URL}/health`);
    return response.ok();
  },

  async healthWorker(request: APIRequestContext): Promise<boolean> {
    const workerOneResponse = await request.get(`${urls.WORKER_1_URL}/health`);
    const workerTwoResponse = await request.get(`${urls.WORKER_2_URL}/health`);
    return workerOneResponse.ok() && workerTwoResponse.ok();
  },

  /**
   * @param searchPath is relative to the e2e-tests directory. Defaults to Banananation.
   */
  async uploadJarFile(
    request: APIRequestContext,
    searchPath: string = '../examples/banananation/build/libs/*',
  ): Promise<number> {
    const absoluteSearchPath = resolve(searchPath);
    const [jarPath = 'ERROR_JAR_NOT_FOUND'] = FastGlob.sync(absoluteSearchPath);

    const buffer = createReadStream(jarPath);
    const name = basename(jarPath);
    const mimeType = 'application/java-archive';
    const multipart = { buffer, mimeType, name };
    const response = await request.post(`${urls.GATEWAY_URL}/file`, { multipart });

    if (response.ok()) {
      const json = await response.json();

      if (json?.id !== undefined) {
        const { id: jar_id } = json;
        return jar_id;
      } else if (json?.success === false) {
        console.error(json);
        throw new Error(json.message);
      } else {
        throw new Error('An unexpected error ocurred');
      }
    } else {
      throw new Error(response.statusText());
    }
  },

  async createPlan(request: APIRequestContext, model: CreatePlanInput): Promise<number> {
    const data = await req.hasura(request, gql.CREATE_PLAN, { plan: model });
    const { insert_plan_one } = data;
    const { id: plan_id } = insert_plan_one;
    return plan_id;
  },

  async createSimulation(request: APIRequestContext, simulationInput: SimulationCreation): Promise<number> {
    const data = await req.hasura(request, gql.CREATE_SIMULATION, { simulation: simulationInput });
    const { insert_simulation_one } = data;
    const { id: simulation_id } = insert_simulation_one;
    return simulation_id;
  },

  async insertSchedulingGoal(request: APIRequestContext, schedulingInput: SchedulingGoalInsertInput) {
    const data = await req.hasura(request, gql.CREATE_SCHEDULING_GOAL, {goal: schedulingInput});
    const { insert_scheduling_goal_one } = data;
    const { id: goal_id } = insert_scheduling_goal_one;
    return goal_id
  },

  async insertSchedulingSpecification(request: APIRequestContext, specificationInput: SchedulingSpecInsertInput) {
    const data = await req.hasura(request, gql.INSERT_SCHEDULING_SPECIFICATION, {scheduling_spec: specificationInput});
    const { insert_scheduling_specification_one } = data;
    const { id: spec_id } = insert_scheduling_specification_one;
    return spec_id
  },

  async getPlanRevision(request: APIRequestContext, id: number): Promise<number | null> {
    const data = await req.hasura(request, gql.GET_PLAN_REVISION, { id: id });
    const { plan } = data;
    const { revision } = plan;
    return revision;
  },

  async schedule(request: APIRequestContext, specificationId: number): Promise<SchedulingResponse> {
    const data = await req.hasura(request, gql.SCHEDULE, {
      specificationId: specificationId,
    });
    const { schedule } = data;
    return schedule;
  },

  async createSchedulingSpecGoal(request: APIRequestContext, spec_goal: SchedulingSpecGoalInsertInput): Promise<void> {
    const data = await req.hasura(request, gql.CREATE_SCHEDULING_SPEC_GOAL, {spec_goal});
    const { insert_scheduling_specification_goals_one } = data;
    const {
      goal_id: goal_id,
      priority : priority,
      specification_id: specification_id} = insert_scheduling_specification_goals_one
    return specification_id;
  },

  async deletePlan(request: APIRequestContext, id: number){
    const data = await req.hasura(request, gql.DELETE_PLAN, { id: id })
    const { deletePlan } = data;
    const {id : deletedPlan} = deletePlan;
    return deletedPlan;
  },

  async getPlan(request: APIRequestContext, id: number): Promise<Plan>{
    const data = await req.hasura<{ plan: Plan }>(request, gql.GET_PLAN, { id: id });
    const { plan } = data;
    const startTime = new Date(plan.startTime);
    return {
      ...plan,
      activity_directives: plan.activity_directives.map((activity: any) => toActivity(activity, startTime)),
      endTime: time.getDoyTimeFromDuration(startTime, plan.duration),
      startTime: time.getDoyTime(startTime),
    };
  },

};
/**
 * Converts any activity to an Activity.
 */
export function toActivity(activity: any, startTime: Date): ActivityDirective {
  return {
    arguments: activity.arguments,
    children: [],
    duration: 0,
    id: activity.id,
    parent: null,
    startTime: time.getDoyTimeFromDuration(startTime, activity.startOffset),
    type: activity.type,
  };
}

export default req;
